"""Service for predicting energy usage."""
import datetime

from estimenergy.const import MetricPeriod
from estimenergy.helpers import get_days_in_month, get_days_in_year
from estimenergy.models.config.config import Config
from estimenergy.models.device_config import DeviceConfig


class PredictionService:
    """Service for predicting energy usage."""

    device_config: DeviceConfig
    config: Config

    def __init__(self, device_config: DeviceConfig, config: Config):
        self.device_config = device_config
        self.config = config

    async def calculate_cost(
        self,
        metric_period: MetricPeriod,
        energy: float,
        date: datetime.date = datetime.date.today(),
    ) -> float:
        """Calculate the cost."""

        if metric_period == MetricPeriod.DAY:
            days_in_month = get_days_in_month(date.month, date.year)
            base_cost = self.device_config.base_cost_per_month / days_in_month
        elif metric_period == MetricPeriod.MONTH:
            base_cost = self.device_config.base_cost_per_month
        elif metric_period == MetricPeriod.YEAR:
            base_cost = self.device_config.base_cost_per_month * 12
        else:
            base_cost = 0

        energy_cost = energy * self.device_config.cost_per_kwh
        cost = base_cost + energy_cost

        return cost

    async def calculate_cost_difference(
        self,
        metric_period: MetricPeriod,
        energy: float,
        date: datetime.date = datetime.date.today(),
    ) -> float:
        """Calculate the cost difference."""

        if metric_period == MetricPeriod.DAY:
            days_in_month = get_days_in_month(date.month, date.year)
            payment = self.device_config.payment_per_month / days_in_month
        elif metric_period == MetricPeriod.MONTH:
            payment = self.device_config.payment_per_month
        elif metric_period == MetricPeriod.YEAR:
            payment = self.device_config.payment_per_month * 12
        else:
            return 0

        cost = await self.calculate_cost(metric_period, energy, date)
        cost_difference = payment - cost

        return cost_difference

    async def calculate_accuracy(
        self,
        metric_period: MetricPeriod,
        accuracy_sum: float,
        date: datetime.date = datetime.date.today(),
    ) -> float:
        """Calculate the accuracy."""

        if metric_period == MetricPeriod.MONTH:
            day_count = get_days_in_month(date.year, date.month)
        elif metric_period == MetricPeriod.YEAR:
            day_count = get_days_in_year(date.year)
        else:
            return 0

        total_accuracy = accuracy_sum / day_count

        return total_accuracy

    async def predict_energy(
        self,
        metric_period: MetricPeriod,
        energy: float,
        accurate_day_count: int,
        date: datetime.date = datetime.date.today(),
    ) -> float:
        """Predict the energy."""

        if metric_period == MetricPeriod.MONTH:
            day_count = get_days_in_month(date.year, date.month)
        elif metric_period == MetricPeriod.YEAR:
            day_count = get_days_in_year(date.year)
        else:
            return 0

        if accurate_day_count == 0:
            return 0
        mean_energy_per_day = energy / accurate_day_count
        predicted_energy = mean_energy_per_day * day_count

        return predicted_energy

    async def predict_cost(
        self,
        metric_period: MetricPeriod,
        energy: float,
        accurate_day_count: int,
        date: datetime.date = datetime.date.today(),
    ) -> float:
        """Predict the cost."""

        predicted_energy = await self.predict_energy(
            metric_period, energy, accurate_day_count, date
        )
        predicted_cost = await self.calculate_cost(
            metric_period, predicted_energy, date
        )

        return predicted_cost

    async def predict_cost_difference(
        self,
        metric_period: MetricPeriod,
        energy: float,
        accurate_day_count: int,
        date: datetime.date = datetime.date.today(),
    ) -> float:
        """Predict the cost difference."""

        predicted_energy = await self.predict_energy(
            metric_period, energy, accurate_day_count, date
        )
        predicted_cost_difference = await self.calculate_cost_difference(
            metric_period, predicted_energy, date
        )

        return predicted_cost_difference



import datetime

from prometheus_client import Gauge, Metric as PrometheusMetric

from estimenergy.const import METRICS, Metric
from estimenergy.models.config.config import Config
from estimenergy.models.device_config import DeviceConfig
from estimenergy.prometheus import metric_registry
from estimenergy.services.data_service import DataService
from estimenergy.services.sql_service import SqlService


class PrometheusService(DataService):
    sql_service: SqlService
    gauges: dict[Metric, Gauge]

    def __init__(
        self, sql_service: SqlService, device_config: DeviceConfig, config: Config
    ):
        super().__init__(device_config, config)

        self.sql_service = sql_service
        self.gauges = {}

        for metric in METRICS:
            try:
                self.gauges[metric] = metric.create_gauge(metric_registry).labels(
                    name=device_config.name
                )
            except ValueError:
                pass

    @property
    def supported_metrics(self):
        return []

    async def _last(
        self,
        metric: Metric,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ) -> float:
        _ = metric
        _ = value_dt

        return 0

    async def _write(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        _ = metric
        _ = value
        _ = value_dt

    async def update(
        self,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        _ = value_dt

        for metric in METRICS:
            last_value = await self.sql_service.last(metric, value_dt)

            self.gauges[metric].set(last_value)



from typing import Optional

from fastapi import APIRouter, HTTPException
from sqlmodel import Session, select

from estimenergy.db import db_engine
from estimenergy.devices import device_registry
from estimenergy.models.year import Year

year_router = APIRouter(prefix="/year", tags=["year"])


@year_router.get(
    "",
    response_model=list[Year],
    responses={
        404: {"description": "Device not found"},
    },
    operation_id="get_years",
)
async def get_years(device_name: Optional[str] = None):
    """Get all years."""

    if device_name is not None and not await device_registry.device_exists(device_name):
        raise HTTPException(status_code=404, detail="Device not found")

    with Session(db_engine) as session:
        if device_name is None:
            years = session.exec(select(Year)).all()
        else:
            years = session.exec(
                select(Year).where(Year.device_name == device_name)
            ).all()
        return years


@year_router.get(
    "/{id}",
    response_model=Year,
    responses={
        404: {"description": "Year not found"},
    },
    operation_id="get_year",
)
async def get_year(year_id: int):
    """Get a year."""

    with Session(db_engine) as session:
        year = session.exec(select(Year).where(Year.id == year_id)).first()
        if year is None:
            raise HTTPException(status_code=404, detail="Year not found")
        return year



"""Logging configuration for estimenergy."""
import logging
from logging import config as logging_config

from estimenergy.config import config
from estimenergy.const import LOGGING_CONFIG

LOGGING_CONFIG["handlers"]["file"]["filename"] = config.logging_config.log_path
LOGGING_CONFIG["loggers"]["uvicorn.error"]["level"] = config.logging_config.log_level
LOGGING_CONFIG["loggers"]["uvicorn.access"]["level"] = config.logging_config.log_level
LOGGING_CONFIG["loggers"]["estimenergy"]["level"] = config.logging_config.log_level

logging_config.dictConfig(LOGGING_CONFIG)

logger = logging.getLogger("estimenergy")



"""Helper functions."""


def get_days_in_month(month: int, year: int):
    """Return the number of days in a given month and year."""

    if month == 2:
        if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0):
            return 29

        return 28

    if month in {4, 6, 9, 11}:
        return 30

    return 31


def get_days_in_year(year: int):
    """Return the number of days in a given year."""

    if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0):
        return 366

    return 365



"""Abstract class for all devices."""
from abc import ABC, abstractmethod
import datetime

from estimenergy.const import Metric
from estimenergy.models.config.config import Config
from estimenergy.models.device_config import DeviceConfig, DeviceConfigRead
from estimenergy.services.data_service import DataService
from estimenergy.services.influx_service import InfluxService
from estimenergy.services.prometheus_service import PrometheusService
from estimenergy.services.sql_service import SqlService


class BaseDevice(ABC):
    """Abstract class for all devices."""

    config: Config
    device_config: DeviceConfig
    data_services: list[DataService]

    def __init__(self, device_config: DeviceConfig, config: Config):
        self.data_services = []
        self.config = config
        self.device_config = device_config

        sql_service = SqlService(self.device_config, config)
        self.data_services.append(sql_service)

        prometheus_service = PrometheusService(sql_service, self.device_config, config)
        self.data_services.append(prometheus_service)

        if config.influx_config:
            influx_service = InfluxService(self.device_config, config)
            self.data_services.append(influx_service)

    @property
    @abstractmethod
    def provided_metrics(self) -> list[Metric]:
        """Return a list of metrics provided by this device."""

    @abstractmethod
    async def stop(self):
        """Stop the device."""

    @abstractmethod
    async def start(self):
        """Start the device."""

    async def increment(
        self,
        metric: Metric,
        value: float,
        date: datetime.datetime = datetime.datetime.now(),
    ):
        """Increment a metric in the database."""

        if metric not in self.provided_metrics:
            raise ValueError(f"Metric {metric} not provided by this device.")

        for data_service in self.data_services:
            await data_service.increment(metric, value, date)

    async def decrement(
        self,
        metric: Metric,
        value: float,
        date: datetime.datetime = datetime.datetime.now(),
    ):
        """Decrement a metric in the database."""

        if metric not in self.provided_metrics:
            raise ValueError(f"Metric {metric} not provided by this device.")

        for data_service in self.data_services:
            await data_service.decrement(metric, value, date)

    async def write(
        self,
        metric: Metric,
        value: float,
        date: datetime.datetime = datetime.datetime.now(),
    ):
        """Write a metric to the database."""

        if metric not in self.provided_metrics:
            raise ValueError(f"Metric {metric} not provided by this device.")

        for data_service in self.data_services:
            await data_service.write(metric, value, date)

    async def update(
        self,
        date: datetime.datetime = datetime.datetime.now(),
    ):
        """Calculate metrics based on other metrics."""

        for data_service in self.data_services:
            await data_service.update(date)



"""Database module for EstimEnergy application."""
from sqlmodel import Session, SQLModel, create_engine

from estimenergy.config import config
from estimenergy.log import logger

db_engine = create_engine(config.sql_config.url)


def create_db():
    """Create the database for the EstimEnergy application."""

    logger.info("Creating database...")

    with Session(db_engine) as session:
        SQLModel.metadata.create_all(db_engine)
        session.commit()


def drop_db():
    """Drop all tables."""

    logger.info("Dropping all tables...")

    with Session(db_engine) as session:
        SQLModel.metadata.drop_all(db_engine)
        session.commit()



from typing import Optional

from sqlmodel import Session, select

from estimenergy.config import config
from estimenergy.const import DeviceType
from estimenergy.db import db_engine
from estimenergy.devices.base_device import BaseDevice
from estimenergy.devices.device_error import DeviceError
from estimenergy.devices.glow_device import GlowDevice
from estimenergy.log import logger
from estimenergy.models.device_config import DeviceConfig


class DeviceRegistry:
    devices: list[BaseDevice] = []

    async def initialize(self):
        """Initialize the DeviceRegistry class."""

        logger.info("Initializing device registry...")

        with Session(db_engine) as session:
            device_configs = session.exec(select(DeviceConfig)).all()

        for device_config in device_configs:
            device = await self.create_device(device_config)

            if device_config.is_active:
                try:
                    await device.start()
                except DeviceError:
                    continue

    async def device_exists(self, device_name: str) -> bool:
        """Check if a device exists."""

        return await self.get_device(device_name) is not None

    async def get_device(self, device_name: str) -> Optional[BaseDevice]:
        """Get a device by name."""

        for device in self.devices:
            if device.device_config.name == device_name:
                return device

        return None

    async def create_device(self, device_config: DeviceConfig) -> BaseDevice:
        """Create a device for the EstimEnergy application."""

        logger.info(f"Creating device {device_config.name}...")

        with Session(db_engine) as session:
            session.add(device_config)
            session.commit()
            session.refresh(device_config)

        if device_config.type == DeviceType.GLOW:
            device = GlowDevice(device_config, config)
            self.devices.append(device)

        return device

    async def update_device(
        self, device: BaseDevice, device_config: DeviceConfig
    ) -> BaseDevice:
        """Update a device for the EstimEnergy application."""

        logger.info(f"Updating device {device.device_config.name}...")

        if device.device_config.is_active:
            await device.stop()

        with Session(db_engine) as session:
            db_device_config: DeviceConfig = session.exec(
                select(DeviceConfig).where(
                    DeviceConfig.name == device.device_config.name
                )
            ).one()

            db_device_config.name = device_config.name
            db_device_config.type = device_config.type
            db_device_config.host = device_config.host
            db_device_config.port = device_config.port
            db_device_config.password = device_config.password
            db_device_config.cost_per_kwh = device_config.cost_per_kwh
            db_device_config.base_cost_per_month = device_config.base_cost_per_month
            db_device_config.payment_per_month = device_config.payment_per_month
            db_device_config.billing_month = device_config.billing_month
            db_device_config.min_accuracy = device_config.min_accuracy

            session.add(db_device_config)
            session.commit()
            session.refresh(db_device_config)

        self.devices.remove(device)

        if device_config.type == DeviceType.GLOW:
            device = GlowDevice(db_device_config, config)
            self.devices.append(device)

        await device.start()

        return device

    async def delete_device(self, device: BaseDevice):
        """Delete a device for the EstimEnergy application."""

        logger.info(f"Deleting device {device.device_config.name}...")

        if device.device_config.is_active:
            await device.stop()

        with Session(db_engine) as session:
            session.delete(device.device_config)
            session.commit()

        self.devices.remove(device)


device_registry = DeviceRegistry()



"""Device config models."""
from sqlmodel import Field, SQLModel

from estimenergy.const import DeviceType


class BaseDeviceConfig(SQLModel):
    """Base device config."""

    name: str = Field(default=None, primary_key=True)
    type: DeviceType = Field(default=DeviceType.GLOW)
    host: str = Field(default=None)
    port: int = Field(default=None)
    cost_per_kwh: float = Field(default=0)
    base_cost_per_month: float = Field(default=0)
    payment_per_month: float = Field(default=0)
    billing_month: int = Field(default=1)
    min_accuracy: float = Field(default=0)
    is_active: bool = Field(default=True)
    is_connected: bool = Field(default=False)


class DeviceConfig(BaseDeviceConfig, table=True):
    """Device config with secrets."""

    password: str = Field(default="")


class DeviceConfigRead(BaseDeviceConfig):
    """Device config without secrets."""



"""Entrypoint for the EstimEnergy application."""
import json

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
import uvicorn

from estimenergy.config import config
from estimenergy.const import API_PREFIX
from estimenergy.db import create_db
from estimenergy.devices import device_registry
from estimenergy.log import logger
from estimenergy.prometheus import instrumentator
from estimenergy.routers.day_router import day_router
from estimenergy.routers.device_router import device_router
from estimenergy.routers.month_router import month_router
from estimenergy.routers.total_router import total_router
from estimenergy.routers.year_router import year_router

app = FastAPI(
    title="EstimEnergy",
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=config.networking_config.cors_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

instrumentator.instrument(app, "estimenergy")
instrumentator.expose(app, include_in_schema=True)

app.include_router(device_router, prefix=API_PREFIX)
app.include_router(day_router, prefix=API_PREFIX)
app.include_router(month_router, prefix=API_PREFIX)
app.include_router(year_router, prefix=API_PREFIX)
app.include_router(total_router, prefix=API_PREFIX)


def start():
    """Start the EstimEnergy application."""

    uvicorn.run(
        "estimenergy.main:app",
        host=config.networking_config.host,
        port=config.networking_config.port,
        reload=config.dev_config.reload,
    )


def generate_openapi():
    """Generate the OpenAPI schema for the EstimEnergy application."""

    with open("openapi.json", "w", encoding="utf-8") as openapi_file:
        dump = json.dumps(app.openapi(), indent=2)
        openapi_file.write(dump)


@app.on_event("startup")
async def startup():
    """Application startup event."""

    logger.info("Starting EstimEnergy application.")

    create_db()

    await device_registry.initialize()

    logger.info("EstimEnergy application started.")



from typing import Optional

from fastapi import APIRouter, HTTPException
from sqlmodel import Session, select

from estimenergy.db import db_engine
from estimenergy.devices import device_registry
from estimenergy.models.total import Total

total_router = APIRouter(prefix="/total", tags=["total"])


@total_router.get(
    "",
    response_model=list[Total],
    responses={
        404: {"description": "Device not found"},
    },
    operation_id="get_totals",
)
async def get_totals(device_name: Optional[str] = None):
    """Get all totals."""

    if device_name is not None and not await device_registry.device_exists(device_name):
        raise HTTPException(status_code=404, detail="Device not found")

    with Session(db_engine) as session:
        if device_name is None:
            totals = session.exec(select(Total)).all()
        else:
            totals = session.exec(
                select(Total).where(Total.device_name == device_name)
            ).all()
        return totals


@total_router.get(
    "/{id}",
    response_model=Total,
    responses={
        404: {"description": "Total not found"},
    },
    operation_id="get_total",
)
async def get_total(total_id: int):
    """Get a total."""

    with Session(db_engine) as session:
        total = session.exec(select(Total).where(Total.id == total_id)).first()
        if total is None:
            raise HTTPException(status_code=404, detail="Total not found")
        return total



from influxdb_client import InfluxDBClient

from estimenergy.config import config

if config.influx_config:
    influx_client = InfluxDBClient(
        url=config.influx_config.url,
        token=config.influx_config.token.get_secret_value(),
        org=config.influx_config.org,
    )
else:
    influx_client = None



from pydantic import BaseModel

from estimenergy.const import DEFAULT_HOST, DEFAULT_PORT


class NetworkingConfig(BaseModel):
    host: str = DEFAULT_HOST
    port: int = DEFAULT_PORT
    cors_origins: list = [
        "http://localhost",
        "http://localhost:12321",
    ]



"""Day router."""
import datetime
from typing import Optional

from fastapi import APIRouter, HTTPException
from fastapi.responses import JSONResponse
from sqlmodel import Session, select

from estimenergy.const import RESPONSE_DAY_NOT_FOUND, RESPONSE_DEVICE_NOT_FOUND
from estimenergy.db import db_engine
from estimenergy.devices import device_registry
from estimenergy.devices.base_device import BaseDevice
from estimenergy.models.day import Day, DayCreate, DayRead
from estimenergy.models.message import Message

day_router = APIRouter(prefix="/day", tags=["day"])


@day_router.get(
    "",
    response_model=list[DayRead],
    responses={
        404: {"model": Message},
    },
    operation_id="get_days",
)
async def get_days(device_name: Optional[str] = None):
    """Get all days."""

    if device_name is not None and not await device_registry.device_exists(device_name):
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    with Session(db_engine) as session:
        if device_name is None:
            days = session.exec(select(Day)).all()
        else:
            days = session.exec(select(Day).where(Day.device_name == device_name)).all()
        return days


@day_router.get(
    "/{day_id}",
    response_model=DayRead,
    responses={
        404: {"model": Message},
    },
    operation_id="get_day",
)
async def get_day(day_id: int):
    """Get a day."""

    with Session(db_engine) as session:
        day = session.exec(select(Day).where(Day.id == day_id)).first()

    if day is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DAY_NOT_FOUND},
        )

    return day


@day_router.post(
    "",
    response_model=DayRead,
    responses={
        404: {"model": Message},
    },
    operation_id="create_day",
)
async def create_day(day: DayCreate):
    """Create a day."""

    device: Optional[BaseDevice] = await device_registry.get_device(day.device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    with Session(db_engine) as session:
        db_day = Day.from_orm(day)
        session.add(db_day)
        session.commit()
        session.refresh(db_day)

        date = datetime.datetime(db_day.date.year, db_day.date.month, db_day.date.day)
        await device.update(date)
        session.commit()

        session.refresh(db_day)

        return db_day


@day_router.put(
    "/{day_id}",
    response_model=DayRead,
    responses={
        404: {"model": Message},
    },
    operation_id="update_day",
)
async def update_day(day_id: int, day: DayCreate):
    """Update a day."""

    device: Optional[BaseDevice] = await device_registry.get_device(day.device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    with Session(db_engine) as session:
        db_day = session.exec(select(Day).where(Day.id == day_id)).first()

        if db_day is None:
            return JSONResponse(
                status_code=404,
                content={"message": RESPONSE_DAY_NOT_FOUND},
            )

        db_day.energy = day.energy
        db_day.date = day.date
        db_day.accuracy = day.accuracy

        session.add(db_day)
        session.commit()
        session.refresh(db_day)

        date = datetime.datetime(db_day.date.year, db_day.date.month, db_day.date.day)
        await device.update(date)
        session.commit()

        session.refresh(db_day)

        return db_day


@day_router.delete(
    "/{day_id}",
    response_model=DayRead,
    responses={
        404: {"model": Message},
    },
    operation_id="delete_day",
)
async def delete_day(day_id: int):
    """Delete a day."""

    with Session(db_engine) as session:
        day = session.exec(select(Day).where(Day.id == day_id)).first()

        if day is None:
            return JSONResponse(
                status_code=404,
                content={"message": RESPONSE_DAY_NOT_FOUND},
            )

        session.delete(day)
        session.commit()

        return day



"""InfluxDB data service."""
import datetime

from influxdb_client import Point, WritePrecision
from influxdb_client.client.flux_table import FluxRecord, FluxTable

from estimenergy.const import Metric, MetricPeriod, MetricType
from estimenergy.influx import influx_client
from estimenergy.services.data_service import DataService


class InfluxService(DataService):
    """InfluxDB data service."""

    @property
    def supported_metrics(self) -> list[Metric]:
        """Return a list of metrics supported by this service."""
        return [
            Metric(
                MetricType.ENERGY,
                MetricPeriod.TOTAL,
                False,
                False,
            ),
            Metric(
                MetricType.COST,
                MetricPeriod.TOTAL,
                False,
                False,
            ),
            Metric(
                MetricType.POWER,
                MetricPeriod.TOTAL,
                False,
                False,
            ),
        ]

    async def _last(
        self,
        metric: Metric,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ) -> float:
        """Get the last value for a metric type."""

        if self.config.influx_config is None:
            return 0

        if metric.metric_period != MetricPeriod.TOTAL:
            return 0

        if metric.metric_type not in [MetricType.ENERGY, MetricType.COST]:
            return 0

        timestamp = value_dt.strftime("%Y-%m-%dT%H:%M:%SZ")

        result: list[FluxTable] = influx_client.query_api().query(
            f'from(bucket:"{self.config.influx_config.bucket}") |> range(start: 0) |> filter(fn: (r) => r._measurement == "energy" and r._field == "{metric.metric_type.value[0]}") |> filter(fn: (r) => r._time >= {timestamp}) |> sort(columns: ["_time"], desc: false) |> limit(n: 1)'
        )

        if len(result) == 0:
            return 0

        last_data: FluxTable = result[0]
        records: list[FluxRecord] = last_data.records

        if len(records) == 0:
            return 0

        last_record: FluxRecord = records[0]
        return last_record.get_value()

    async def _write(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Write a metric to the database."""

        if self.config.influx_config is None:
            return

        if metric.metric_period != MetricPeriod.TOTAL:
            return

        if metric.metric_type not in [MetricType.ENERGY, MetricType.COST]:
            return

        point = (
            Point("energy")
            .tag("device", self.device_config.name)
            .field(metric.metric_type.value[0], value)
            .time(value_dt, WritePrecision.MS)
        )

        influx_client.write_api().write(
            bucket=self.config.influx_config.bucket, record=point
        )

    async def update(
        self,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Update the database with the latest values."""
        _ = value_dt



from prometheus_client import CollectorRegistry
from prometheus_fastapi_instrumentator import Instrumentator

metric_registry = CollectorRegistry()
instrumentator = Instrumentator(
    should_group_status_codes=False,
    should_ignore_untemplated=True,
    should_respect_env_var=False,
    should_instrument_requests_inprogress=True,
    excluded_handlers=[".*admin.*", "/metrics"],
    inprogress_name="inprogress",
    inprogress_labels=True,
    registry=metric_registry,
)



from typing import Optional

from fastapi import APIRouter, HTTPException
from sqlmodel import Session, select

from estimenergy.db import db_engine
from estimenergy.devices import device_registry
from estimenergy.models.month import Month

month_router = APIRouter(prefix="/month", tags=["month"])


@month_router.get(
    "",
    response_model=list[Month],
    responses={
        404: {"description": "Device not found"},
    },
    operation_id="get_months",
)
async def get_months(device_name: Optional[str] = None):
    """Get all months."""

    if device_name is not None and not await device_registry.device_exists(device_name):
        raise HTTPException(status_code=404, detail="Device not found")

    with Session(db_engine) as session:
        if device_name is None:
            months = session.exec(select(Month)).all()
        else:
            months = session.exec(
                select(Month).where(Month.device_name == device_name)
            ).all()
        return months


@month_router.get(
    "/{id}",
    response_model=Month,
    responses={
        404: {"description": "Month not found"},
    },
    operation_id="get_month",
)
async def get_month(month_id: int):
    """Get a month."""

    with Session(db_engine) as session:
        month = session.exec(select(Month).where(Month.id == month_id)).first()
        if month is None:
            raise HTTPException(status_code=404, detail="Month not found")
        return month



"""The main config class."""
from typing import Optional

from pydantic import BaseModel, Field
import yaml

from estimenergy.models.config.dev_config import DevConfig
from estimenergy.models.config.influx_config import InfluxConfig
from estimenergy.models.config.logging_config import LoggingConfig
from estimenergy.models.config.networking_config import NetworkingConfig
from estimenergy.models.config.sql_config import SqlConfig
from estimenergy.models.device_config import DeviceConfig


class Config(BaseModel):
    """The main config class."""

    dev_config: DevConfig = Field(DevConfig(), alias="dev")
    networking_config: NetworkingConfig = Field(NetworkingConfig(), alias="networking")
    logging_config: LoggingConfig = Field(LoggingConfig(), alias="logging")
    sql_config: SqlConfig = Field(SqlConfig(), alias="db")
    influx_config: Optional[InfluxConfig] = Field(alias="influxdb")

    @classmethod
    def from_file(cls, config_path: str):
        """Load the config from a file."""

        with open(config_path, "r", encoding="utf-8") as config_file:
            config_dict: dict = yaml.safe_load(config_file)

        config = cls.parse_obj(config_dict)
        return config

    @classmethod
    def with_location(cls, path: str):
        """Create default config with base location."""

        config = Config()
        config.sql_config.url = f"sqlite:///{path}/estimenergy.db"
        config.logging_config.log_path = f"{path}/estimenergy.log"

        return config



from fastapi import APIRouter, HTTPException
from fastapi.responses import JSONResponse
from sqlmodel import Session

from estimenergy.const import (
    RESPONSE_DEVICE_DELETED,
    RESPONSE_DEVICE_FAILED_TO_START,
    RESPONSE_DEVICE_NOT_FOUND,
)
from estimenergy.db import db_engine
from estimenergy.devices import device_registry
from estimenergy.devices.device_error import DeviceError
from estimenergy.models.device_config import DeviceConfig, DeviceConfigRead
from estimenergy.models.message import Message

device_router = APIRouter(prefix="/device", tags=["device"])


@device_router.post(
    "",
    response_model=DeviceConfigRead,
    operation_id="create_device",
    responses={
        500: {"model": Message},
    },
)
async def create_device(device_config: DeviceConfig):
    """Create a new device."""

    device = await device_registry.create_device(device_config)

    try:
        await device.start()
    except DeviceError:
        return JSONResponse(
            status_code=500,
            content={"message": RESPONSE_DEVICE_FAILED_TO_START},
        )

    return device.device_config


@device_router.get(
    "/{device_name}",
    response_model=DeviceConfigRead,
    operation_id="get_device",
    responses={
        404: {"model": Message},
    },
)
async def get_device(device_name: str):
    """Get a device."""

    device = await device_registry.get_device(device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    return device.device_config


@device_router.get(
    "", response_model=list[DeviceConfigRead], operation_id="get_devices"
)
async def get_devices():
    """Get all devices."""

    device_configs = [device.device_config for device in device_registry.devices]
    return device_configs


@device_router.put(
    "/{device_name}",
    response_model=DeviceConfigRead,
    operation_id="update_device",
    responses={
        404: {"model": Message},
    },
)
async def update_device(device_name: str, device_config: DeviceConfig):
    """Update a device."""

    device = await device_registry.get_device(device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    device = await device_registry.update_device(device, device_config)

    return device.device_config


@device_router.delete(
    "/{device_name}",
    operation_id="delete_device",
    responses={
        404: {"model": Message},
        201: {"model": Message},
    },
)
async def delete_device(device_name: str):
    """Delete a device."""

    device = await device_registry.get_device(device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    await device_registry.delete_device(device)

    return JSONResponse(
        status_code=201,
        content={"message": RESPONSE_DEVICE_DELETED},
    )


@device_router.post(
    "/{device_name}/start",
    response_model=DeviceConfigRead,
    operation_id="start_device",
    responses={
        404: {"model": Message},
        500: {"model": Message},
    },
)
async def start_device(device_name: str):
    """Start a device."""

    device = await device_registry.get_device(device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    try:
        await device.start()
    except DeviceError:
        return JSONResponse(
            status_code=500,
            content={"message": RESPONSE_DEVICE_FAILED_TO_START},
        )

    return device.device_config


@device_router.post(
    "/{device_name}/stop",
    response_model=DeviceConfigRead,
    operation_id="stop_device",
    responses={
        404: {"model": Message},
    },
)
async def stop_device(device_name: str):
    """Stop a device."""

    device = await device_registry.get_device(device_name)

    if device is None:
        return JSONResponse(
            status_code=404,
            content={"message": RESPONSE_DEVICE_NOT_FOUND},
        )

    await device.stop()

    return device.device_config



"""Configuration module for estimenergy."""
import os

from dotenv import load_dotenv

from estimenergy.const import DEFAULT_CONFIG_PATH
from estimenergy.models.config.config import Config

load_dotenv()

config_path = os.getenv("CONFIG_PATH", DEFAULT_CONFIG_PATH)

if os.path.exists(config_path):
    config = Config.from_file(config_path)
else:
    config_base_path = os.path.dirname(config_path)
    config = Config.with_location(config_base_path)



"""Day model."""
import datetime
from typing import Optional

from sqlmodel import Field, SQLModel


class DayBase(SQLModel):
    """Day base model."""

    device_name: str = Field(default=None, index=True)
    date: datetime.date = Field(default=None, index=True)
    energy: float = Field(default=0)
    accuracy: float = Field(default=0)


class DayBaseCalc(DayBase):
    """Day base model with calculated fields."""

    cost: float = Field(default=0)
    cost_difference: float = Field(default=0)


class Day(DayBaseCalc, table=True):
    """Day model."""

    id: Optional[int] = Field(default=None, primary_key=True)
    year_id: Optional[int] = Field(default=None, foreign_key="year.id")
    month_id: Optional[int] = Field(default=None, foreign_key="month.id")


class DayCreate(DayBase):
    """Day create model."""


class DayRead(DayBaseCalc):
    """Day read model."""

    id: int



"""Total model."""
from typing import Optional

from sqlmodel import Field, SQLModel


class Total(SQLModel, table=True):
    """Total model."""

    id: Optional[int] = Field(default=None, primary_key=True)
    device_name: str = Field(default=None, index=True)
    energy: float = Field(default=0)
    cost: float = Field(default=0)



import requests


class EstimEnergyClient:
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def get_device(self, name: str):
        url = f"http://{self.host}:{self.port}/api/device/{name}"
        response = requests.get(url, timeout=5)
        return response.json()

    def get_devices(self):
        url = f"http://{self.host}:{self.port}/api/device"
        response = requests.get(url, timeout=5)
        return response.json()

    def get_metrics(self):
        url = f"http://{self.host}:{self.port}/metrics"
        response = requests.get(url, timeout=5)
        return response.text



from enum import Enum

from prometheus_client import Gauge

API_PREFIX = "/api"

DEFAULT_CONFIG_PATH = "config.yml"
DEFAULT_LOG_PATH = "estimenergy.log"
DEFAULT_LOG_LEVEL = "INFO"

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 12321

DEFAULT_INFLUXDB_HOST = "127.0.0.1"
DEFAULT_INFLUXDB_PORT = 8086
DEFAULT_INFLUXDB_USERNAME = "estimenergy"

SENSOR_TYPE_JSON = "json"
SENSOR_TYPE_FRIENDLY_NAME = "friendly_name"
SENSOR_TYPE_UNIQUE_ID = "unique_id"

RESPONSE_DEVICE_NOT_FOUND = "Device not found"
RESPONSE_DEVICE_FAILED_TO_START = "Device failed to start"
RESPONSE_DEVICE_DELETED = "Device deleted"
RESPONSE_DAY_NOT_FOUND = "Day not found"


class DeviceType(Enum):
    GLOW = "glow"


class MetricPeriod(Enum):
    DAY = ("day", "Daily")
    MONTH = ("month", "Monthly")
    YEAR = ("year", "Yearly")
    TOTAL = ("total", "Total")


class MetricType(Enum):
    COST = ("cost", "Cost")
    COST_DIFFERENCE = ("cost_difference", "Cost Difference")
    ENERGY = ("energy", "Energy")
    ACCURACY = ("accuracy", "Accuracy")
    POWER = ("power", "Power")


class Metric:
    def __init__(
        self,
        metric_type: MetricType,
        metric_period: MetricPeriod,
        is_predicted: bool,
        is_raw: bool,
    ):
        self.metric_type = metric_type
        self.metric_period = metric_period
        self.is_predicted = is_predicted
        self.is_raw = is_raw

    @property
    def key(self) -> str:
        return f"{self.metric_type.value[0]}{'_predicted' if self.is_predicted else ''}{'_raw' if self.is_raw else ''}"

    @property
    def metric_key(self) -> str:
        return f"estimenergy_{self.metric_period.value[0]}_{self.metric_type.value[0]}{'_predicted' if self.is_predicted else ''}{'_raw' if self.is_raw else ''}"

    @property
    def friendly_name(self) -> str:
        return f"{self.metric_period.value[1]} {self.metric_type.value[1]} {'(Predicted)' if self.is_predicted else ''} {'(Raw)' if self.is_raw else ''}"

    def create_gauge(self, registry) -> Gauge:
        return Gauge(
            f"{self.metric_key}",
            f"EstimEnergy {self.friendly_name}",
            ["name"],
            registry=registry,
        )

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Metric):
            return (
                self.metric_type == __value.metric_type
                and self.metric_period == __value.metric_period
                and self.is_predicted == __value.is_predicted
                and self.is_raw == __value.is_raw
            )
        return False

    def __hash__(self) -> int:
        return hash(
            (self.metric_type, self.metric_period, self.is_predicted, self.is_raw)
        )


METRICS = [
    Metric(MetricType.ENERGY, MetricPeriod.DAY, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.MONTH, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.YEAR, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.TOTAL, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.MONTH, True, False),
    Metric(MetricType.ENERGY, MetricPeriod.YEAR, True, False),
    Metric(MetricType.ENERGY, MetricPeriod.MONTH, True, True),
    Metric(MetricType.ENERGY, MetricPeriod.YEAR, True, True),
    Metric(MetricType.COST, MetricPeriod.DAY, False, False),
    Metric(MetricType.COST, MetricPeriod.MONTH, False, False),
    Metric(MetricType.COST, MetricPeriod.YEAR, False, False),
    Metric(MetricType.COST, MetricPeriod.TOTAL, False, False),
    Metric(MetricType.COST, MetricPeriod.MONTH, True, False),
    Metric(MetricType.COST, MetricPeriod.YEAR, True, False),
    Metric(MetricType.COST, MetricPeriod.MONTH, True, True),
    Metric(MetricType.COST, MetricPeriod.YEAR, True, True),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.DAY, False, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.MONTH, False, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.YEAR, False, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.MONTH, True, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.YEAR, True, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.MONTH, True, True),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.YEAR, True, True),
    Metric(MetricType.ACCURACY, MetricPeriod.DAY, False, False),
    Metric(MetricType.ACCURACY, MetricPeriod.MONTH, False, False),
    Metric(MetricType.ACCURACY, MetricPeriod.YEAR, False, False),
]

JSON_NAME = "name"
JSON_HOST = "host"
JSON_PORT = "port"
JSON_COST_PER_KWH = "cost_per_kwh"
JSON_BASE_COST_PER_MONTH = "base_cost_per_month"
JSON_PAYMENT_PER_MONTH = "payment_per_month"
JSON_MIN_HOUR = "min_hour"
JSON_MAX_HOUR = "max_hour"
JSON_MAX_INCOMPLETE_DAYS = "max_incomplete_days"
JSON_BILLING_MONTH = "billing_month"
JSON_DATA = "data"

LOGGING_CONFIG: dict = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(asctime)-25s %(name)-30s %(levelname)-8s %(message)s",
        },
        "file": {
            "format": "%(asctime)-25s %(name)-30s %(levelname)-8s %(message)s",
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
        "file": {
            "formatter": "file",
            "class": "logging.FileHandler",
            "filename": "estimenergy.log",
        },
    },
    "loggers": {
        "uvicorn.error": {
            "level": "INFO",
            "handlers": ["default", "file"],
            "propagate": False,
        },
        "uvicorn.access": {
            "level": "INFO",
            "handlers": ["access", "file"],
            "propagate": False,
        },
        "estimenergy": {
            "level": "INFO",
            "handlers": ["default", "file"],
            "propagate": False,
        },
    },
}



"""Year model."""
import datetime
from typing import Optional

from sqlmodel import Field, SQLModel


class Year(SQLModel, table=True):
    """Year model."""

    id: Optional[int] = Field(default=None, primary_key=True)
    device_name: str = Field(default=None, index=True)
    date: datetime.date = Field(default=None, index=True)
    energy: float = Field(default=0)
    cost: float = Field(default=0)
    cost_difference: float = Field(default=0)
    accuracy: float = Field(default=0)
    energy_predicted_raw: float = Field(default=0)
    energy_predicted: float = Field(default=0)
    cost_predicted_raw: float = Field(default=0)
    cost_predicted: float = Field(default=0)
    cost_difference_predicted_raw: float = Field(default=0)
    cost_difference_predicted: float = Field(default=0)



"""Abstract class for all data services."""
from abc import ABC, abstractmethod
import datetime

from estimenergy.const import Metric
from estimenergy.models.config.config import Config
from estimenergy.models.device_config import DeviceConfig


class DataService(ABC):
    """Abstract class for all data services."""

    device_config: DeviceConfig
    config: Config

    def __init__(self, device_config: DeviceConfig, config: Config):
        self.device_config = device_config
        self.config = config

    async def last(
        self,
        metric: Metric,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ) -> float:
        """Return the last value for a metric."""

        if metric not in self.supported_metrics:
            return 0

        return await self._last(metric, value_dt)

    async def write(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Write a metric to the database."""

        if metric not in self.supported_metrics:
            return

        await self._write(metric, value, value_dt)

    async def increment(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Increment a metric in the database."""

        if metric not in self.supported_metrics:
            return

        last_value = await self.last(metric, value_dt)

        await self.write(metric, last_value + value, value_dt)

    async def decrement(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Decrement a metric in the database."""

        if metric not in self.supported_metrics:
            return

        last_value = await self.last(metric, value_dt)

        await self.write(metric, last_value - value, value_dt)

    @property
    @abstractmethod
    def supported_metrics(self) -> list[Metric]:
        """Return a list of metrics supported by this service."""

    @abstractmethod
    async def _last(
        self,
        metric: Metric,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ) -> float:
        """Get the last value for a metric."""

    @abstractmethod
    async def _write(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Write a metric to the database."""

    @abstractmethod
    async def update(
        self,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Update metrics."""



"""Service for writing and reading data from the SQL database."""
import datetime
import traceback

from sqlalchemy import extract
from sqlmodel import Session, select
from sqlmodel.sql.expression import SelectOfScalar

from estimenergy.const import METRICS, Metric, MetricPeriod, MetricType
from estimenergy.db import db_engine
from estimenergy.models.config.config import Config
from estimenergy.models.day import Day
from estimenergy.models.device_config import DeviceConfig
from estimenergy.models.month import Month
from estimenergy.models.total import Total
from estimenergy.models.year import Year
from estimenergy.services.data_service import DataService
from estimenergy.services.prediction_service import PredictionService


class SqlService(DataService):
    """Service for writing and reading data from the SQL database."""

    prediction_service: PredictionService

    def __init__(self, device_config: DeviceConfig, config: Config):
        super().__init__(device_config, config)

        self.prediction_service = PredictionService(device_config, config)

    @property
    def supported_metrics(self) -> list[Metric]:
        return [
            metric
            for metric in METRICS
            if not metric.is_raw and metric.metric_type != MetricType.POWER
        ]

    async def _last(
        self,
        metric: Metric,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ) -> float:
        """Get a metric value."""

        date = value_dt.date()
        row = self.get_or_create_row(metric.metric_period, date)
        value = getattr(row, metric.key, 0)

        return value

    async def _write(
        self,
        metric: Metric,
        value: float,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Write a metric value."""

        date = value_dt.date()
        row = self.get_or_create_row(metric.metric_period, date)
        setattr(row, metric.key, value)

        with Session(db_engine) as session:
            session.add(row)
            session.commit()

    async def update(
        self,
        value_dt: datetime.datetime = datetime.datetime.now(),
    ):
        """Update all metrics."""

        date = value_dt.date()
        await self.update_day(date)
        await self.update_month(date)
        await self.update_year(date)
        await self.update_total()

    async def update_day(
        self,
        date: datetime.date = datetime.date.today(),
    ):
        """Update the day metrics."""

        metric_period = MetricPeriod.DAY

        with Session(db_engine) as session:
            day = self.get_or_create_row(MetricPeriod.DAY, date)

            if not isinstance(day, Day):
                raise TypeError("day is not of type Day")

            day.cost = await self.prediction_service.calculate_cost(
                metric_period,
                day.energy,
                day.date,
            )
            day.cost_difference = (
                await self.prediction_service.calculate_cost_difference(
                    metric_period,
                    day.energy,
                    day.date,
                )
            )

            session.add(day)
            session.commit()

    async def update_month(
        self,
        date: datetime.date = datetime.date.today(),
    ):
        """Update the month metrics."""

        await self.update_accumulative(MetricPeriod.MONTH, date)

    async def update_year(
        self,
        date: datetime.date = datetime.date.today(),
    ):
        """Update the year metrics."""

        await self.update_accumulative(MetricPeriod.YEAR, date)

    async def update_accumulative(
        self,
        metric_period: MetricPeriod,
        date: datetime.date = datetime.date.today(),
    ):
        """Update accumulative metrics."""

        with Session(db_engine) as session:
            row = self.get_or_create_row(metric_period, date)

            if metric_period == MetricPeriod.MONTH:
                recorded_days = session.exec(
                    select(Day).where(
                        Day.month_id == row.id,
                    )
                ).all()
                accurate_days = session.exec(
                    select(Day).where(
                        Day.month_id == row.id,
                        Day.accuracy >= self.device_config.min_accuracy,
                    )
                ).all()
            elif metric_period == MetricPeriod.YEAR:
                recorded_days = session.exec(
                    select(Day).where(
                        Day.year_id == row.id,
                    )
                ).all()
                accurate_days = session.exec(
                    select(Day).where(
                        Day.year_id == row.id,
                        Day.accuracy >= self.device_config.min_accuracy,
                    )
                ).all()
            else:
                raise ValueError(f"Invalid metric period: {metric_period}")

            recorded_day_count = len(recorded_days)
            accurate_day_count = len(accurate_days)

            energy_raw = sum(day.energy for day in recorded_days)
            energy = sum(day.energy for day in accurate_days)
            energy_predicted_raw = await self.prediction_service.predict_energy(
                metric_period,
                energy_raw,
                recorded_day_count,
                date,
            )
            energy_predicted = await self.prediction_service.predict_energy(
                metric_period,
                energy,
                accurate_day_count,
                date,
            )

            cost = await self.prediction_service.calculate_cost(
                metric_period,
                energy_raw,
                date,
            )
            cost_predicted_raw = await self.prediction_service.predict_cost(
                metric_period,
                energy_raw,
                recorded_day_count,
                date,
            )
            cost_predicted = await self.prediction_service.predict_cost(
                metric_period,
                energy,
                accurate_day_count,
                date,
            )

            cost_difference = await self.prediction_service.calculate_cost_difference(
                metric_period,
                energy_raw,
                date,
            )
            cost_difference_predicted_raw = (
                await self.prediction_service.predict_cost_difference(
                    metric_period,
                    energy_raw,
                    recorded_day_count,
                    date,
                )
            )
            cost_difference_predicted = (
                await self.prediction_service.predict_cost_difference(
                    metric_period,
                    energy,
                    accurate_day_count,
                    date,
                )
            )

            accuracy_sum = sum(day.accuracy for day in accurate_days)
            accuracy = await self.prediction_service.calculate_accuracy(
                metric_period,
                accuracy_sum,
                date,
            )

            row.energy = energy_raw
            row.energy_predicted_raw = energy_predicted_raw
            row.energy_predicted = energy_predicted
            row.cost = cost
            row.cost_predicted_raw = cost_predicted_raw
            row.cost_predicted = cost_predicted
            row.cost_difference = cost_difference
            row.cost_difference_predicted_raw = cost_difference_predicted_raw
            row.cost_difference_predicted = cost_difference_predicted
            row.accuracy = accuracy

            session.add(row)
            session.commit()

    async def update_total(self):
        """Update the total metrics."""

        metric_period = MetricPeriod.TOTAL

        with Session(db_engine) as session:
            total = self.get_or_create_row(metric_period)
            days = session.exec(select(Day)).all()
            energy = sum(day.energy for day in days)

            cost = await self.prediction_service.calculate_cost(
                metric_period,
                energy,
            )

            total.energy = energy
            total.cost = cost

            session.add(total)
            session.commit()

    def get_or_create_row(
        self,
        metric_period: MetricPeriod,
        date: datetime.date = datetime.date.today(),
    ) -> Day | Month | Year | Total:
        """Get or create a row in the database."""

        row = self.get_row(metric_period, date)

        if row is None:
            row = self.create_row(metric_period, date)

        return row

    def create_row(
        self,
        metric_period: MetricPeriod,
        date: datetime.date = datetime.date.today(),
    ) -> Day | Month | Year | Total:
        """Create a row in the database."""

        if metric_period == MetricPeriod.DAY:
            return self.create_day(date)

        elif metric_period == MetricPeriod.MONTH:
            return self.create_month(date)

        elif metric_period == MetricPeriod.YEAR:
            return self.create_year(date)

        elif metric_period == MetricPeriod.TOTAL:
            return self.create_total()

        else:
            raise ValueError(f"Unknown metric period {metric_period}")

    def create_day(
        self,
        date: datetime.date = datetime.date.today(),
    ) -> Day:
        """Create a day in the database."""

        with Session(db_engine, expire_on_commit=False) as session:
            row = Day(device_name=self.device_config.name, date=date)
            session.add(row)
            session.commit()
            month = session.exec(
                select(Month).where(
                    extract("month", Month.date) == date.month,
                    extract("year", Month.date) == date.year,
                    Month.device_name == self.device_config.name,
                )
            ).first()
            if month is not None:
                row.month_id = month.id

            session.add(row)
            session.commit()

        return row

    def create_month(
        self,
        date: datetime.date = datetime.date.today(),
    ) -> Month:
        """Create a month in the database."""

        with Session(db_engine, expire_on_commit=False) as session:
            date = date.replace(day=1)
            row = Month(device_name=self.device_config.name, date=date)
            session.add(row)
            session.commit()
            days = session.exec(
                select(Day).where(
                    extract("month", Day.date) == date.month,
                    extract("year", Day.date) == date.year,
                    Day.device_name == self.device_config.name,
                )
            ).all()
            for day in days:
                day.month_id = row.id
                session.add(day)
            year = session.exec(
                select(Year).where(
                    extract("year", Year.date) == date.year
                    and extract("month", Year.date) <= date.month
                    or extract("year", Year.date) == date.year + 1
                    and extract("month", Year.date) > date.month,
                    Year.device_name == self.device_config.name,
                )
            ).first()
            if year is not None:
                row.year_id = year.id

            session.commit()

        return row

    def create_year(
        self,
        date: datetime.date = datetime.date.today(),
    ) -> Year:
        """Create a year in the database."""

        with Session(db_engine, expire_on_commit=False) as session:
            date = date.replace(
                year=date.year - 1
                if date.month <= self.device_config.billing_month
                else date.year,
                month=self.device_config.billing_month,
                day=1,
            )
            row = Year(device_name=self.device_config.name, date=date)
            session.add(row)
            session.commit()
            months = session.exec(
                select(Month).where(
                    extract("year", Month.date) == date.year
                    and extract("month", Month.date) >= date.month
                    or extract("year", Month.date) == date.year + 1
                    and extract("month", Month.date) < date.month,
                    Month.device_name == self.device_config.name,
                )
            ).all()
            for month in months:
                month.year_id = row.id
                session.add(month)
            days = session.exec(
                select(Day).where(
                    extract("year", Day.date) == date.year
                    and extract("month", Day.date) >= date.month
                    or extract("year", Day.date) == date.year + 1
                    and extract("month", Day.date) < date.month,
                    Day.device_name == self.device_config.name,
                )
            ).all()
            for day in days:
                day.year_id = row.id
                session.add(day)

            session.commit()

        return row

    def create_total(self) -> Total:
        """Create a total in the database."""

        with Session(db_engine, expire_on_commit=False) as session:
            row = Total(device_name=self.device_config.name)

            session.add(row)
            session.commit()

        return row

    def get_rows(
        self,
        metric_period: MetricPeriod,
        date: datetime.date = datetime.date.today(),
    ) -> list[Day] | list[Month] | list[Year] | list[Total]:
        """Get rows from the database."""

        query: SelectOfScalar[Day] | SelectOfScalar[Month] | SelectOfScalar[
            Year
        ] | SelectOfScalar[Total]

        with Session(db_engine, expire_on_commit=False) as session:
            if metric_period == MetricPeriod.DAY:
                query = select(Day).where(
                    Day.date == date,
                    Day.device_name == self.device_config.name,
                )
            elif metric_period == MetricPeriod.MONTH:
                date = date.replace(day=1)
                query = select(Month).where(
                    Month.date == date,
                    Month.device_name == self.device_config.name,
                )
            elif metric_period == MetricPeriod.YEAR:
                date = date.replace(
                    year=date.year - 1
                    if date.month <= self.device_config.billing_month
                    else date.year,
                    month=self.device_config.billing_month,
                    day=1,
                )
                query = select(Year).where(
                    Year.date == date,
                    Year.device_name == self.device_config.name,
                )
            elif metric_period == MetricPeriod.TOTAL:
                query = select(Total).where(
                    Total.device_name == self.device_config.name,
                )
            else:
                raise ValueError(f"Unknown metric period {metric_period}")

            return session.exec(query).all()

    def get_row(
        self,
        metric_period: MetricPeriod,
        date: datetime.date = datetime.date.today(),
    ) -> Day | Month | Year | Total | None:
        """Get a row from the database."""

        rows = self.get_rows(metric_period, date)

        if len(rows) == 0:
            return None

        return rows[0]



"""Home Assistant Glow device."""
import asyncio
import datetime
from typing import Optional

from aioesphomeapi import (
    APIClient,
    APIConnectionError,
    DeviceInfo,
    InvalidEncryptionKeyAPIError,
    ReconnectLogic,
    RequiresEncryptionAPIError,
    ResolveAPIError,
    SensorState,
)
from sqlmodel import Session
from zeroconf import Zeroconf

from estimenergy.const import Metric, MetricPeriod, MetricType
from estimenergy.db import db_engine
from estimenergy.devices import BaseDevice
from estimenergy.devices.device_error import DeviceError
from estimenergy.log import logger
from estimenergy.models.config.config import Config
from estimenergy.models.device_config import DeviceConfig


class GlowDevice(BaseDevice):
    """Home Assistant Glow device."""

    zeroconf: Zeroconf
    api: APIClient
    reconnect_logic: Optional[ReconnectLogic] = None
    last_kwh: Optional[float] = None
    last_time: Optional[datetime.datetime] = None

    def __init__(self, device_config: DeviceConfig, config: Config):
        """Initialize the Glow device."""

        super().__init__(device_config, config)

        self.zeroconf = Zeroconf()
        self.api = APIClient(
            self.device_config.host,
            self.device_config.port,
            self.device_config.password,
            zeroconf_instance=self.zeroconf,
        )

    @property
    def provided_metrics(self) -> list[Metric]:
        return [
            Metric(MetricType.ENERGY, MetricPeriod.DAY, False, False),
            Metric(MetricType.ACCURACY, MetricPeriod.DAY, False, False),
            Metric(MetricType.POWER, MetricPeriod.TOTAL, False, False),
        ]

    async def start(self):
        """Start the device."""
        with Session(db_engine) as session:
            self.device_config.is_active = True
            session.add(self.device_config)
            session.commit()
            session.refresh(self.device_config)

        if not await self.can_connect():
            logger.error(f"Unable to login to {self.device_config.name}")
            raise DeviceError(f"Unable to login to {self.device_config.name}")

        event = asyncio.Event()

        async def on_connect():
            logger.info(f"Connected to ESPHome Device {self.device_config.name}")
            event.set()

            with Session(db_engine) as session:
                self.device_config.is_connected = True
                session.add(self.device_config)
                session.commit()
                session.refresh(self.device_config)

            await self.api.subscribe_states(self.__state_changed)

        async def on_connect_error():
            with Session(db_engine) as session:
                self.device_config.is_connected = False
                session.add(self.device_config)
                session.commit()
                session.refresh(self.device_config)

            raise DeviceError(f"Unable to connect to {self.device_config.name}")

        async def on_disconnect():
            with Session(db_engine) as session:
                self.device_config.is_connected = False
                session.add(self.device_config)
                session.commit()
                session.refresh(self.device_config)

        self.reconnect_logic = ReconnectLogic(
            client=self.api,
            on_connect=on_connect,
            on_disconnect=on_disconnect,
            zeroconf_instance=self.zeroconf,
            name=self.device_config.name,
            on_connect_error=on_connect_error,
        )

        await self.reconnect_logic.start()

        await event.wait()

    async def stop(self):
        """Stop the device."""

        logger.info(f"Stopping ESPHome Device {self.device_config.name}")

        with Session(db_engine) as session:
            self.device_config.is_active = False
            session.add(self.device_config)
            session.commit()
            session.refresh(self.device_config)

        await self.api.disconnect(force=True)
        if self.reconnect_logic is not None:
            await self.reconnect_logic.stop()
        self.zeroconf.close()

    async def can_connect(self) -> bool:
        """Check if we can connect to the device."""
        try:
            await self.api.connect(login=True)
            return True
        except (
            ResolveAPIError,
            APIConnectionError,
            InvalidEncryptionKeyAPIError,
            RequiresEncryptionAPIError,
        ):
            return False
        finally:
            await self.api.disconnect(force=True)

    def __state_changed(self, state: SensorState):
        loop = asyncio.get_event_loop()

        if state.key == 2274151077:
            loop.create_task(self.__on_power_changed(state.state))
            return

        if state.key == 2690257735:
            loop.create_task(self.__on_total_kwh_changed(state.state))

    async def __on_total_kwh_changed(self, value: float):
        if self.last_kwh is None or self.last_time is None:
            self.last_kwh = value
            self.last_time = datetime.datetime.now()
            return

        if value < self.last_kwh:
            self.last_kwh = value
            logger.warning("Detected a reset of the total kWh counter.")
            return

        time = datetime.datetime.now()

        kwh_increase = value - self.last_kwh
        time_increase_us = (time - self.last_time).microseconds
        us_per_day = 1000 * 1000 * 60 * 60 * 24
        accuracy_increase = time_increase_us / us_per_day

        logger.debug(
            f"Detected {kwh_increase} kWh increase in {time_increase_us} us for device {self.device_config.name}."
        )

        await self.increment(
            Metric(MetricType.ENERGY, MetricPeriod.DAY, False, False), kwh_increase
        )
        await self.increment(
            Metric(MetricType.ACCURACY, MetricPeriod.DAY, False, False),
            accuracy_increase,
        )

        await self.update()

        self.last_kwh = value

    async def __on_power_changed(self, value: float):
        await self.write(
            Metric(MetricType.POWER, MetricPeriod.TOTAL, False, False), value
        )



from pydantic import BaseModel

from estimenergy.const import DEFAULT_LOG_LEVEL, DEFAULT_LOG_PATH


class LoggingConfig(BaseModel):
    log_path: str = DEFAULT_LOG_PATH
    log_level: str = DEFAULT_LOG_LEVEL
